from dataclasses import dataclass
from typing import List, Dict

from cyanprintsdk.domain.core.answer import Answer
from cyanprintsdk.domain.core.cyan import Cyan


@dataclass
class ExtensionAnswerInput:
    answers: List[Answer]
    deterministic_state: List[Dict[str, str]]
    prev_answers: List[Answer]
    prev_cyan: Cyan
    prev_extension_answers: Dict[str, List[Answer]]
    prev_extension_cyans: Dict[str, Cyan]


@dataclass
class ExtensionValidateInput:
    answers: List[Answer]
    deterministic_state: List[Dict[str, str]]
    prev_answers: List[Answer]
    prev_cyan: Cyan
    prev_extension_answers: Dict[str, List[Answer]]
    prev_extension_cyans: Dict[str, Cyan]
    validate: str



from cyanprintsdk.domain.core.cyan_script import ICyanPlugin
from cyanprintsdk.domain.core.cyan_script_model import CyanPluginInput
from cyanprintsdk.domain.plugin.input import PluginInput
from cyanprintsdk.domain.plugin.output import PluginOutput


class PluginService:
    def __init__(self, plugin: ICyanPlugin):
        self._plugin = plugin  # Leading underscore indicates a "private" attribute

    async def plug(self, i: PluginInput) -> PluginOutput:
        directory = i.directory
        config = i.config
        return await self._plugin.plugin(CyanPluginInput(directory=directory, config=config))



from cyanprintsdk.api.plugin.req import PluginReq
from cyanprintsdk.api.plugin.res import PluginRes
from cyanprintsdk.domain.plugin.input import PluginInput
from cyanprintsdk.domain.plugin.output import PluginOutput


class PluginMapper:
    @staticmethod
    def to_domain(req: PluginReq) -> PluginInput:
        return PluginInput(directory=req.directory, config=req.config)

    @staticmethod
    def to_res(res: PluginOutput) -> PluginRes:
        return PluginRes(output_dir=res.directory)



from typing import Callable, Awaitable

from cyanprintsdk.domain.core.cyan_script import ICyanProcessor
from cyanprintsdk.domain.core.cyan_script_model import CyanProcessorInput
from cyanprintsdk.domain.core.fs.cyan_fs_helper import CyanFileHelper
from cyanprintsdk.domain.processor.output import ProcessorOutput

LambdaProcessorFn = Callable[[CyanProcessorInput, CyanFileHelper], Awaitable[ProcessorOutput]]


class LambdaProcessor(ICyanProcessor):
    def __init__(self, f: LambdaProcessorFn):
        self._f: LambdaProcessorFn = f

    async def process(self, i: CyanProcessorInput, file_helper: CyanFileHelper) -> ProcessorOutput:
        return await self._f(i, file_helper)



from dataclasses import dataclass
from typing import Union, List


@dataclass
class StringArrayAnswer:
    answer: List[str]


@dataclass
class StringAnswer:
    answer: str


@dataclass
class BoolAnswer:
    answer: bool


Answer = Union[StringAnswer, StringArrayAnswer, BoolAnswer]


def is_string_answer(a: Answer) -> bool:
    return isinstance(a, StringAnswer)


def is_string_array_answer(a: Answer) -> bool:
    return isinstance(a, StringArrayAnswer)


def is_bool_answer(a: Answer) -> bool:
    return isinstance(a, BoolAnswer)



import pprint

from aiohttp import web
from pydantic import ValidationError

from cyanprintsdk.api.extension.fn import LambdaExtensionFn, LambdaExtension
from cyanprintsdk.api.extension.mapper import ExtensionInputMapper, ExtensionOutputMapper
from cyanprintsdk.api.extension.req import ExtensionAnswerReq, ExtensionValidateReq
from cyanprintsdk.api.extension.res import ExtensionRes, ExtensionValidRes
from cyanprintsdk.api.plugin.fn import LambdaPluginFn, LambdaPlugin
from cyanprintsdk.api.plugin.mapper import PluginMapper
from cyanprintsdk.api.plugin.req import PluginReq
from cyanprintsdk.api.plugin.res import PluginRes
from cyanprintsdk.api.processor.fn import LambdaProcessorFn, LambdaProcessor
from cyanprintsdk.api.processor.mapper import ProcessorMapper
from cyanprintsdk.api.processor.req import ProcessorReq
from cyanprintsdk.api.processor.res import ProcessorRes
from cyanprintsdk.api.template.fn import LambdaTemplate, LambdaTemplateFn
from cyanprintsdk.api.template.mapper import TemplateInputMapper, TemplateOutputMapper
from cyanprintsdk.api.template.req import TemplateAnswerReq, TemplateValidateReq
from cyanprintsdk.api.template.res import TemplateRes, TemplateValidRes
from cyanprintsdk.domain.core.cyan_script import ICyanPlugin, ICyanProcessor, ICyanTemplate, ICyanExtension
from cyanprintsdk.domain.extension.input import ExtensionAnswerInput, ExtensionValidateInput
from cyanprintsdk.domain.extension.output import ExtensionOutput
from cyanprintsdk.domain.extension.service import ExtensionService
from cyanprintsdk.domain.plugin.input import PluginInput
from cyanprintsdk.domain.plugin.output import PluginOutput
from cyanprintsdk.domain.plugin.service import PluginService
from cyanprintsdk.domain.processor.input import ProcessorInput
from cyanprintsdk.domain.processor.output import ProcessorOutput
from cyanprintsdk.domain.processor.service import ProcessorService
from cyanprintsdk.domain.template.input import TemplateInput, TemplateValidateInput
from cyanprintsdk.domain.template.output import TemplateOutput
from cyanprintsdk.domain.template.service import TemplateService


async def health_check(request):
    return web.json_response({
        "Message": "OK",
        "Status": "OK",
    })


def start_plugin_with_fn(f: LambdaPluginFn):
    start_plugin(LambdaPlugin(f))


def start_plugin(plugin: ICyanPlugin):
    app = web.Application()

    plugin_service = PluginService(plugin)

    async def plug(request):
        try:
            json = await request.json()
            req = PluginReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: PluginInput = PluginMapper.to_domain(req)
        o: PluginOutput = await plugin_service.plug(i)
        res: PluginRes = PluginMapper.to_res(o)

        return web.json_response(res.model_dump())

    app.add_routes([
        web.get("/", health_check),
        web.post('/api/plug', plug),

    ])

    web.run_app(app, port=5552)


def start_processor_with_fn(f: LambdaProcessorFn):
    start_processor(LambdaProcessor(f))


def start_processor(processor: ICyanProcessor):
    app = web.Application()

    proc_service = ProcessorService(processor)

    async def process(request):
        try:
            json = await request.json()
            req = ProcessorReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: ProcessorInput = ProcessorMapper.to_domain(req)
        o: ProcessorOutput = await proc_service.process(i)
        res: ProcessorRes = ProcessorMapper.to_res(o)

        return web.json_response(res.model_dump())

    app.add_routes([
        web.get("/", health_check),
        web.post('/api/process', process),
    ])

    web.run_app(app, port=5551)


def start_template_with_fn(f: LambdaTemplateFn):
    start_template(LambdaTemplate(f))


def start_template(template: ICyanTemplate):
    app = web.Application()

    template_service = TemplateService(template)

    async def template_answer(request):
        try:
            json = await request.json()
            print(json)
            req = TemplateAnswerReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: TemplateInput = TemplateInputMapper.answer_to_domain(req)
        o: TemplateOutput = await template_service.template(i)
        res: TemplateRes = TemplateOutputMapper.to_resp(o)

        res_model = res.model_dump(by_alias=True)

        return web.json_response(res_model)

    async def template_validate(request):
        try:
            json = await request.json()
            req = TemplateValidateReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: TemplateValidateInput = TemplateInputMapper.validate_to_domain(req)
        o: str = await template_service.validate(i)
        res: TemplateValidRes = TemplateValidRes(valid=o)

        res_model = res.model_dump(by_alias=True)

        return web.json_response(res_model)

    app.add_routes([
        web.get("/", health_check),
        web.post('/api/template/init', template_answer),
        web.post('/api/template/validate', template_validate),
    ])

    web.run_app(app, port=5550)


def start_extension_with_fn(f: LambdaExtensionFn):
    start_extension(LambdaExtension(f))


def start_extension(extension: ICyanExtension):
    app = web.Application()

    ext_service = ExtensionService(extension)

    async def ext_answer(request):
        try:
            json = await request.json()
            req = ExtensionAnswerReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: ExtensionAnswerInput = ExtensionInputMapper.extension_answer_to_domain(req)
        o: ExtensionOutput = await ext_service.extend(i)
        res: ExtensionRes = ExtensionOutputMapper.to_resp(o)

        return web.json_response(res.model_dump())

    async def ext_validate(request):
        try:
            json = await request.json()
            req = ExtensionValidateReq(**json)
            pprint.pprint(req)
        except ValidationError as e:
            print(e)
            return web.json_response({'error': str(e)}, status=400)

        # translate to domain
        i: ExtensionValidateInput = ExtensionInputMapper.extension_validate_to_domain(req)
        o: str = await ext_service.validate(i)
        res: ExtensionValidRes = ExtensionValidRes(valid=o)

        return web.json_response(res.model_dump())

    app.add_routes([
        web.get("/", health_check),
        web.post('/api/extension/init', ext_answer),
        web.post('/api/extension/validate', ext_validate),
    ])

    web.run_app(app, port=5550)



from typing import Optional, List, Dict, Union

from cyanprintsdk.api.base_model import CyanBaseModel
from cyanprintsdk.api.core.cyan_res import CyanRes
from cyanprintsdk.api.core.question_res import QuestionRes


class TemplateValidRes(CyanBaseModel):
    valid: Optional[str]


class TemplateFinalRes(CyanBaseModel):
    cyan: CyanRes
    type: str = "final"


class TemplateQnARes(CyanBaseModel):
    deterministic_state: List[Dict[str, str]]
    question: QuestionRes
    type: str = "questionnaire"


# Union type for TemplateRes
TemplateRes = Union[TemplateQnARes, TemplateFinalRes]



from dataclasses import dataclass
from typing import List, Any

from cyanprintsdk.domain.core.cyan import CyanGlob


@dataclass
class ProcessorInput:
    read_directory: str
    write_directory: str
    globs: List[CyanGlob]
    config: Any



from typing import List, Dict

from cyanprintsdk.api.base_model import CyanBaseModel

from cyanprintsdk.api.core.answer_req import AnswerReq
from cyanprintsdk.api.core.cyan_req import CyanReq


class ExtensionValidateReq(CyanBaseModel):
    answers: List[AnswerReq]
    deterministic_states: List[Dict[str, str]]
    prev_answers: List[AnswerReq]
    prev_cyan: CyanReq
    prev_extension_answers: Dict[str, List[AnswerReq]]
    prev_extension_cyans: Dict[str, CyanReq]
    validate: str


class ExtensionAnswerReq(CyanBaseModel):
    answers: List[AnswerReq]
    deterministic_states: List[Dict[str, str]]
    prev_answers: List[AnswerReq]
    prev_cyan: CyanReq
    prev_extension_answers: Dict[str, List[AnswerReq]]
    prev_extension_cyans: Dict[str, CyanReq]



from abc import abstractmethod, ABC

from cyanprintsdk.domain.core.cyan import Cyan
from cyanprintsdk.domain.core.cyan_script_model import CyanExtensionInput, CyanProcessorInput, CyanPluginInput
from cyanprintsdk.domain.core.deterministic import IDeterminism
from cyanprintsdk.domain.core.fs.cyan_fs_helper import CyanFileHelper
from cyanprintsdk.domain.core.inquirer import IInquirer
from cyanprintsdk.domain.plugin.output import PluginOutput
from cyanprintsdk.domain.processor.output import ProcessorOutput


class ICyanTemplate(ABC):
    @abstractmethod
    async def template(self, inquirer: IInquirer, determinism: IDeterminism) -> Cyan:
        pass


class ICyanExtension(ABC):
    @abstractmethod
    async def extension(self, inquirer: IInquirer, determinism: IDeterminism, i: CyanExtensionInput) -> Cyan:
        pass


class ICyanProcessor(ABC):
    @abstractmethod
    async def process(self, i: CyanProcessorInput, file_helper: CyanFileHelper) -> ProcessorOutput:
        pass


class ICyanPlugin(ABC):
    @abstractmethod
    async def plugin(self, i: CyanPluginInput) -> PluginOutput:
        pass



from typing import List, Union

from cyanprintsdk.api.base_model import CyanBaseModel


class StringArrayAnswerRes(CyanBaseModel):
    answer: List[str]


class StringAnswerRes(CyanBaseModel):
    answer: str


class BoolAnswerRes(CyanBaseModel):
    answer: bool


# Union type for AnswerRes
AnswerRes = Union[StringArrayAnswerRes, StringAnswerRes, BoolAnswerRes]



from cyanprintsdk.api.core.core_mapper import AnswerMapper, CyanMapper, QuestionMapper
from cyanprintsdk.api.extension.req import ExtensionAnswerReq, ExtensionValidateReq
from cyanprintsdk.api.extension.res import ExtensionRes, ExtensionQnARes, ExtensionFinalRes
from cyanprintsdk.domain.extension.input import ExtensionAnswerInput, ExtensionValidateInput
from cyanprintsdk.domain.extension.output import ExtensionOutput, is_final


class ExtensionInputMapper:
    @staticmethod
    def extension_answer_to_domain(req: ExtensionAnswerReq) -> ExtensionAnswerInput:
        answers = [AnswerMapper.to_domain(x) for x in req.answers]
        prev_answers = [AnswerMapper.to_domain(x) for x in req.prev_answers]
        prev_cyan = CyanMapper.cyan_req_to_domain(req.prev_cyan)
        prev_extension_answers = {
            key: [AnswerMapper.to_domain(a) for a in value]
            for key, value in req.prev_extension_answers.items()
        }
        prev_extension_cyans = {
            key: CyanMapper.cyan_req_to_domain(value)
            for key, value in req.prev_extension_cyans.items()
        }

        return ExtensionAnswerInput(
            answers=answers,
            prev_answers=prev_answers,
            deterministic_state=req.deterministic_states,
            prev_cyan=prev_cyan,
            prev_extension_answers=prev_extension_answers,
            prev_extension_cyans=prev_extension_cyans
        )

    @staticmethod
    def extension_validate_to_domain(req: ExtensionValidateReq) -> ExtensionValidateInput:
        return ExtensionValidateInput(
            answers=[AnswerMapper.to_domain(x) for x in req.answers],
            deterministic_state=req.deterministic_states,
            prev_answers=[AnswerMapper.to_domain(x) for x in req.prev_answers],
            prev_cyan=CyanMapper.cyan_req_to_domain(req.prev_cyan),
            prev_extension_answers={
                key: [AnswerMapper.to_domain(a) for a in value]
                for key, value in req.prev_extension_answers.items()
            },
            prev_extension_cyans={
                key: CyanMapper.cyan_req_to_domain(value)
                for key, value in req.prev_extension_cyans.items()
            },
            validate=req.validate
        )


class ExtensionOutputMapper:
    @staticmethod
    def to_resp(output: ExtensionOutput) -> ExtensionRes:
        if is_final(output):
            return ExtensionQnARes(
                cyan=CyanMapper.cyan_to_resp(output.data),
                type="final"
            )
        else:
            return ExtensionFinalRes(
                deterministic_state=output.deterministic_state,
                question=QuestionMapper.question_to_resp(output.question),
                type="questionnaire"
            )



from typing import List, Dict, Callable

from cyanprintsdk.domain.core.deterministic import IDeterminism


class StatelessDeterminism(IDeterminism):
    def __init__(self, states: List[Dict[str, str]], pointer: int):
        self.states = states
        self._pointer = pointer  # Treat _pointer as a private variable

    def get(self, key: str, origin: Callable[[], str]) -> str:
        if self._pointer + 1 >= len(self.states):
            raise RuntimeError("NullReferenceException")

        states = self.states[self._pointer + 1]
        state = states.get(key)

        if state is not None:
            return state

        val = origin()
        states[key] = val
        return val



from dataclasses import dataclass
from typing import List, Dict, Union

from cyanprintsdk.domain.core.cyan import Cyan
from cyanprintsdk.domain.core.question import Question


@dataclass
class TemplateQnAOutput:
    deterministic_state: List[Dict[str, str]]
    question: Question


@dataclass
class TemplateFinalOutput:
    data: Cyan


# Union type to represent either TemplateQnAOutput or TemplateFinalOutput
TemplateOutput = Union[TemplateQnAOutput, TemplateFinalOutput]


def is_final(output: TemplateOutput) -> bool:
    return isinstance(output, TemplateFinalOutput) and output.data is not None


def is_qna(output: TemplateOutput) -> bool:
    return isinstance(output, TemplateQnAOutput) and output.question is not None



from dataclasses import dataclass
from enum import Enum
from typing import Optional, List, Any


@dataclass
class GlobType(Enum):
    Template = 1
    Copy = 2


@dataclass
class CyanGlob:
    root: Optional[str]
    glob: str
    exclude: List[str]
    type: GlobType


@dataclass
class CyanPlugin:
    name: str
    config: Any


@dataclass
class CyanProcessor:
    name: str
    files: List[CyanGlob]
    config: Any


@dataclass
class Cyan:
    processors: List[CyanProcessor]
    plugins: List[CyanPlugin]



from abc import ABC, abstractmethod
from typing import List, Union, Optional

from cyanprintsdk.domain.core.question import CheckboxQ, ConfirmQ, PasswordQ, SelectQ, TextQ, DateQ


class IInquirer(ABC):

    @abstractmethod
    async def checkbox(self, q: Union[CheckboxQ, str], options: Optional[List[str]] = None,
                       desc: Optional[str] = None) -> List[str]:
        pass

    @abstractmethod
    async def confirm(self, q: Union[ConfirmQ, str], desc: Optional[str] = None) -> bool:
        pass

    @abstractmethod
    async def password(self, q: Union[PasswordQ, str], desc: Optional[str] = None) -> str:
        pass

    @abstractmethod
    async def select(self, q: Union[SelectQ, str], options: Optional[List[str]] = None, desc: Optional[str] = None) -> \
            str:
        pass

    @abstractmethod
    async def text(self, q: Union[TextQ, str], desc: Optional[str] = None) -> str:
        pass

    @abstractmethod
    async def date_select(self, q: Union[DateQ, str], desc: Optional[str] = None) -> str:
        pass



from typing import Optional, List, Union

from cyanprintsdk.api.base_model import CyanBaseModel


class ConfirmQuestionRes(CyanBaseModel):
    message: str
    desc: Optional[str] = None
    default: Optional[str] = None
    error_message: Optional[str] = None
    type: str = "confirm"


class DateQuestionRes(CyanBaseModel):
    message: str
    default: Optional[str] = None
    desc: Optional[str] = None
    min_date: Optional[str] = None
    max_date: Optional[str] = None
    type: str = "date"


class CheckboxQuestionRes(CyanBaseModel):
    message: str
    desc: Optional[str] = None
    options: List[str]
    type: str = "checkbox"


class PasswordQuestionRes(CyanBaseModel):
    message: str
    desc: Optional[str] = None
    confirmation: Optional[bool] = None
    type: str = "password"


class SelectQuestionRes(CyanBaseModel):
    message: str
    options: List[str]
    desc: Optional[str] = None
    type: str = "select"


class TextQuestionRes(CyanBaseModel):
    message: str
    default: Optional[str] = None
    desc: Optional[str] = None
    initial: Optional[str] = None
    type: str = "text"


QuestionRes = Union[
    ConfirmQuestionRes, DateQuestionRes, CheckboxQuestionRes, PasswordQuestionRes, SelectQuestionRes, TextQuestionRes]




from typing import List, Any

from cyanprintsdk.api.base_model import CyanBaseModel

from cyanprintsdk.api.core.cyan_req import CyanGlobReq


class ProcessorReq(CyanBaseModel):
    read_dir: str
    write_dir: str
    globs: List[CyanGlobReq]
    config: Any



from typing import Dict, List

from cyanprintsdk.api.base_model import CyanBaseModel

from cyanprintsdk.api.core.answer_req import AnswerReq


class TemplateValidateReq(CyanBaseModel):
    answers: List[AnswerReq]
    deterministic_states: List[Dict[str, str]]
    validate: str


class TemplateAnswerReq(CyanBaseModel):
    answers: List[AnswerReq]
    deterministic_states: List[Dict[str, str]]



from typing import Optional, List, Dict, Union

from cyanprintsdk.api.base_model import CyanBaseModel

from cyanprintsdk.api.core.cyan_res import CyanRes
from cyanprintsdk.api.core.question_res import QuestionRes


class ExtensionValidRes(CyanBaseModel):
    valid: Optional[str]


class ExtensionFinalRes(CyanBaseModel):
    cyan: CyanRes
    type: str = "final"


class ExtensionQnARes(CyanBaseModel):
    deterministic_state: List[Dict[str, str]]
    question: QuestionRes
    type: str = "questionnaire"


ExtensionRes = Union[ExtensionQnARes, ExtensionFinalRes]



from typing import Optional

from cyanprintsdk.domain.core.cyan_script import ICyanExtension
from cyanprintsdk.domain.core.cyan_script_model import CyanExtensionInput
from cyanprintsdk.domain.core.question import Question
from cyanprintsdk.domain.extension.input import ExtensionAnswerInput, ExtensionValidateInput
from cyanprintsdk.domain.extension.output import ExtensionOutput, ExtensionFinalOutput, ExtensionQnAOutput
from cyanprintsdk.domain.service.out_of_answer_error import OutOfAnswerException
from cyanprintsdk.domain.service.stateless_determinism import StatelessDeterminism
from cyanprintsdk.domain.service.stateless_inquirer import StatelessInquirer


class ExtensionService:
    def __init__(self, ext: ICyanExtension):
        self._ext = ext

    async def extend(self, answer: ExtensionAnswerInput) -> ExtensionOutput:
        pointer = -1
        i = StatelessInquirer(answer.answers, pointer)
        d = StatelessDeterminism(answer.deterministic_state, pointer)

        try:
            input_data = CyanExtensionInput(
                prev_answers=answer.prev_answers,
                prev_cyan=answer.prev_cyan,
                prev_extension_answers=answer.prev_extension_answers,
                prev_extension_cyans=answer.prev_extension_cyans,
            )
            r = await self._ext.extension(i, d, input_data)
            return ExtensionFinalOutput(data=r)
        except OutOfAnswerException as e:
            return ExtensionQnAOutput(deterministic_state=d.states, question=e.question)
        except Exception as e:
            raise e

    async def validate(self, answer: ExtensionValidateInput) -> Optional[str]:
        pointer = -1
        i = StatelessInquirer(answer.answers, pointer)
        d = StatelessDeterminism(answer.deterministic_state, pointer)

        try:
            input_data = CyanExtensionInput(
                prev_answers=answer.prev_answers,
                prev_cyan=answer.prev_cyan,
                prev_extension_answers=answer.prev_extension_answers,
                prev_extension_cyans=answer.prev_extension_cyans,
            )
            await self._ext.extension(i, d, input_data)
            raise RuntimeError("Not supposed to reach here for validation!")
        except OutOfAnswerException as e:
            q: Question = e.question
            if q.validate is None:
                return None
            return q.validate(answer.validate)
        except Exception as e:
            raise e

# In Python, no need to explicitly export classes



from typing import List, Union, Optional

from cyanprintsdk.domain.core.answer import Answer, is_string_array_answer, is_bool_answer, is_string_answer
from cyanprintsdk.domain.core.inquirer import IInquirer
from cyanprintsdk.domain.core.question import Question, CheckboxQ, ConfirmQ, PasswordQ, SelectQ, TextQ, DateQ
from cyanprintsdk.domain.service.out_of_answer_error import OutOfAnswerException


class StatelessInquirer(IInquirer):
    def __init__(self, answers: List[Answer], pointer: int):
        self._answers = answers
        self._pointer = pointer

    def _get_answer(self, q: Question) -> Answer:
        if self._pointer == len(self._answers) - 1:
            raise OutOfAnswerException("", q)

        self._pointer += 1
        return self._answers[self._pointer]

    async def checkbox(self, q: Union[CheckboxQ, str], options: Optional[List[str]] = None,
                       desc: Optional[str] = None) -> List[str]:
        if isinstance(q, str):
            if options is None:
                raise ValueError("options cannot be null")
            return await self.checkbox(CheckboxQ(message=q, options=options, desc=desc))

        answer = self._get_answer(q)
        if is_string_array_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: StringArrayAnswer. Got: " + str(type(answer)))

    async def confirm(self, q: Union[ConfirmQ, str], desc: Optional[str] = None) -> bool:
        if isinstance(q, str):
            return await self.confirm(ConfirmQ(message=q, desc=desc))

        answer = self._get_answer(q)
        if is_bool_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: BoolAnswer. Got: " + str(type(answer)))

    async def password(self, q: Union[PasswordQ, str], desc: Optional[str] = None) -> str:
        if isinstance(q, str):
            return await self.password(PasswordQ(message=q, desc=desc))

        answer = self._get_answer(q)
        if is_string_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: StringAnswer. Got: " + str(type(answer)))

    async def select(self, q: Union[SelectQ, str], options: Optional[List[str]] = None,
                     desc: Optional[str] = None) -> str:
        if isinstance(q, str):
            if options is None:
                raise ValueError("options cannot be null")
            return await self.select(SelectQ(message=q, options=options, desc=desc))

        answer = self._get_answer(q)
        if is_string_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: StringAnswer. Got: " + str(type(answer)))

    async def text(self, q: Union[TextQ, str], desc: Optional[str] = None) -> str:
        if isinstance(q, str):
            return await self.text(TextQ(message=q, desc=desc, validate=None, ))

        answer = self._get_answer(q)
        if is_string_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: StringAnswer. Got: " + str(type(answer)))

    async def date_select(self, q: Union[DateQ, str], desc: Optional[str] = None) -> str:
        if isinstance(q, str):
            return await self.date_select(DateQ(message=q, desc=desc))

        answer = self._get_answer(q)
        if is_string_answer(answer):
            return answer.answer

        raise TypeError("Incorrect answer type. Expected: StringAnswer. Got: " + str(type(answer)))



from cyanprintsdk.domain.core.cyan_script import ICyanProcessor
from cyanprintsdk.domain.core.cyan_script_model import CyanProcessorInput
from cyanprintsdk.domain.core.fs.cyan_fs_helper import CyanFileHelper
from cyanprintsdk.domain.processor.input import ProcessorInput
from cyanprintsdk.domain.processor.output import ProcessorOutput


class ProcessorService:
    def __init__(self, processor: ICyanProcessor):
        self._processor = processor  # Leading underscore indicates a "private" attribute

    async def process(self, i: ProcessorInput) -> ProcessorOutput:
        read_directory = i.read_directory
        write_directory = i.write_directory
        globs = i.globs
        config = i.config

        helper = CyanFileHelper(read_directory, write_directory, globs)
        return await self._processor.process(
            CyanProcessorInput(read_dir=read_directory, write_dir=write_directory, globs=globs, config=config),
            helper,
        )



from cyanprintsdk.api.core.answer_req import AnswerReq, is_bool_answer_req, is_string_answer_req, is_string_array_answer_req
from cyanprintsdk.api.core.answer_res import AnswerRes, BoolAnswerRes, StringAnswerRes, StringArrayAnswerRes
from cyanprintsdk.api.core.cyan_req import CyanPluginReq, CyanGlobReq, CyanProcessorReq, CyanReq
from cyanprintsdk.api.core.cyan_res import CyanRes, CyanProcessorRes, CyanPluginRes, CyanGlobRes
from cyanprintsdk.api.core.question_res import QuestionRes, ConfirmQuestionRes, CheckboxQuestionRes, SelectQuestionRes, \
    DateQuestionRes, PasswordQuestionRes, TextQuestionRes
from cyanprintsdk.domain.core.answer import Answer, is_string_array_answer, is_bool_answer, is_string_answer, BoolAnswer, \
    StringAnswer, StringArrayAnswer
from cyanprintsdk.domain.core.cyan import CyanPlugin, CyanGlob, GlobType, CyanProcessor, Cyan
from cyanprintsdk.domain.core.question import Question, ConfirmQ, CheckboxQ, SelectQ, DateQ, PasswordQ, TextQ


class CyanMapper:
    @staticmethod
    def plugin_req_to_domain(req: CyanPluginReq) -> CyanPlugin:
        return CyanPlugin(name=req.name, config=req.config)

    @staticmethod
    def glob_req_to_domain(req: CyanGlobReq) -> CyanGlob:
        if req.type == "template":
            glob_type = GlobType.Template
        elif req.type == "copy":
            glob_type = GlobType.Copy
        else:
            raise ValueError(f"Invalid req.type: {req.type}")

        return CyanGlob(root=req.root, glob=req.glob, exclude=req.exclude, type=glob_type)

    @staticmethod
    def processor_req_to_domain(req: CyanProcessorReq) -> CyanProcessor:
        files = [CyanMapper.glob_req_to_domain(x) for x in req.files]
        return CyanProcessor(name=req.name, config=req.config, files=files)

    @staticmethod
    def cyan_req_to_domain(req: CyanReq) -> Cyan:
        processors = [CyanMapper.processor_req_to_domain(x) for x in req.processors]
        plugins = [CyanMapper.plugin_req_to_domain(x) for x in req.plugins]
        return Cyan(processors=processors, plugins=plugins)

    @staticmethod
    def cyan_to_resp(data: Cyan) -> CyanRes:
        processors = [CyanMapper.processor_to_resp(x) for x in data.processors]
        plugins = [CyanMapper.plugin_to_resp(x) for x in data.plugins]
        return CyanRes(processors=processors, plugins=plugins)

    @staticmethod
    def processor_to_resp(data: CyanProcessor) -> CyanProcessorRes:
        files = [CyanMapper.glob_to_resp(x) for x in data.files]
        return CyanProcessorRes(name=data.name, config=data.config, files=files)

    @staticmethod
    def plugin_to_resp(data: CyanPlugin) -> CyanPluginRes:
        return CyanPluginRes(name=data.name, config=data.config)

    @staticmethod
    def glob_type_to_resp(t: GlobType) -> str:
        if t == GlobType.Template:
            return "template"
        elif t == GlobType.Copy:
            return "copy"
        else:
            raise ValueError(f"Invalid GlobType: {t}")

    @staticmethod
    def glob_to_resp(data: CyanGlob) -> CyanGlobRes:
        return CyanGlobRes(root=data.root, glob=data.glob, exclude=data.exclude,
                           type=CyanMapper.glob_type_to_resp(data.type))


class QuestionMapper:
    @staticmethod
    def question_to_resp(q: Question) -> QuestionRes:
        if isinstance(q, SelectQ):
            return QuestionMapper.select_to_resp(q)
        elif isinstance(q, TextQ):
            return QuestionMapper.text_to_resp(q)
        elif isinstance(q, PasswordQ):
            return QuestionMapper.password_to_resp(q)
        elif isinstance(q, DateQ):
            return QuestionMapper.date_to_resp(q)
        elif isinstance(q, ConfirmQ):
            return QuestionMapper.confirm_to_resp(q)
        elif isinstance(q, CheckboxQ):
            return QuestionMapper.checkbox_to_resp(q)
        else:
            raise ValueError(f"Invalid question type: {q}")

    @staticmethod
    def confirm_to_resp(q: ConfirmQ) -> ConfirmQuestionRes:
        return ConfirmQuestionRes(
            default=q.default,
            message=q.message,
            error_message=q.error_message,
            desc=q.desc,
            type="confirm"
        )

    @staticmethod
    def checkbox_to_resp(q: CheckboxQ) -> CheckboxQuestionRes:
        return CheckboxQuestionRes(
            message=q.message,
            desc=q.desc,
            options=q.options,
            type="checkbox"
        )

    @staticmethod
    def select_to_resp(q: SelectQ) -> SelectQuestionRes:
        return SelectQuestionRes(
            message=q.message,
            desc=q.desc,
            options=q.options,
            type="select"
        )

    @staticmethod
    def text_to_resp(q: TextQ) -> TextQuestionRes:
        return TextQuestionRes(
            message=q.message,
            desc=q.desc,
            default=q.default,
            initial=q.initial,
            type="text"
        )

    @staticmethod
    def password_to_resp(q: PasswordQ) -> PasswordQuestionRes:
        return PasswordQuestionRes(
            message=q.message,
            desc=q.desc,
            confirmation=q.confirmation,
            type="password"
        )

    @staticmethod
    def date_to_resp(q: DateQ) -> DateQuestionRes:
        return DateQuestionRes(
            message=q.message,
            desc=q.desc,
            default=q.default.isoformat() if q.default else None,
            maxDate=q.max_date.isoformat() if q.max_date else None,
            minDate=q.min_date.isoformat() if q.min_date else None,
            type="date"
        )


class AnswerMapper:
    @staticmethod
    def to_domain(req: AnswerReq) -> Answer:
        if is_bool_answer_req(req):
            return BoolAnswer(answer=req.answer)
        elif is_string_answer_req(req):
            return StringAnswer(answer=req.answer)
        elif is_string_array_answer_req(req):
            return StringArrayAnswer(answer=req.answer)
        else:
            raise ValueError(f"Invalid answer request type: {req}")

    @staticmethod
    def to_resp(answer: Answer) -> AnswerRes:
        if is_bool_answer(answer):
            return BoolAnswerRes(answer=answer.answer)
        elif is_string_answer(answer):
            return StringAnswerRes(answer=answer.answer)
        elif is_string_array_answer(answer):
            return StringArrayAnswerRes(answer=answer.answer)
        else:
            raise ValueError(f"Invalid answer type: {answer}")



from typing import Callable, Awaitable

from cyanprintsdk.domain.core.cyan import Cyan
from cyanprintsdk.domain.core.cyan_script import ICyanExtension
from cyanprintsdk.domain.core.cyan_script_model import CyanExtensionInput
from cyanprintsdk.domain.core.deterministic import IDeterminism
from cyanprintsdk.domain.core.inquirer import IInquirer

LambdaExtensionFn = Callable[[IInquirer, IDeterminism, CyanExtensionInput], Awaitable[Cyan]]


class LambdaExtension(ICyanExtension):
    def __init__(self, f: LambdaExtensionFn):
        self._f: LambdaExtensionFn = f

    async def extension(self, inquirer: IInquirer, determinism: IDeterminism, prev: CyanExtensionInput) -> Cyan:
        return await self._f(inquirer, determinism, prev)



from cyanprintsdk.api.core.core_mapper import AnswerMapper, QuestionMapper, CyanMapper
from cyanprintsdk.api.template.req import TemplateAnswerReq, TemplateValidateReq
from cyanprintsdk.api.template.res import TemplateRes, TemplateQnARes, TemplateFinalRes
from cyanprintsdk.domain.template.input import TemplateInput, TemplateValidateInput
from cyanprintsdk.domain.template.output import TemplateOutput, is_qna, is_final


class TemplateInputMapper:
    @staticmethod
    def answer_to_domain(req: TemplateAnswerReq) -> TemplateInput:
        answers = [AnswerMapper.to_domain(x) for x in req.answers]
        return TemplateInput(
            answers=answers,
            deterministic_state=req.deterministic_states,
        )

    @staticmethod
    def validate_to_domain(req: TemplateValidateReq) -> TemplateValidateInput:
        answers = [AnswerMapper.to_domain(x) for x in req.answers]
        return TemplateValidateInput(
            deterministic_state=req.deterministic_states,
            answers=answers,
            validate=req.validate
        )


class TemplateOutputMapper:
    @staticmethod
    def to_resp(output: TemplateOutput) -> TemplateRes:
        if is_qna(output):
            return TemplateQnARes(
                type="questionnaire",
                deterministic_state=output.deterministic_state,
                question=QuestionMapper.question_to_resp(output.question)
            )
        elif is_final(output):
            return TemplateFinalRes(
                cyan=CyanMapper.cyan_to_resp(output.data),
                type="final"
            )
        else:
            raise ValueError(f"Invalid output type {output}")



from dataclasses import dataclass
from typing import List, Dict

from cyanprintsdk.domain.core.answer import Answer


@dataclass
class TemplateInput:
    answers: List[Answer]
    deterministic_state: List[Dict[str, str]]


@dataclass
class TemplateValidateInput:
    answers: List[Answer]
    deterministic_state: List[Dict[str, str]]
    validate: str



from typing import Union, List

from cyanprintsdk.api.base_model import CyanBaseModel


class StringArrayAnswerReq(CyanBaseModel):
    answer: List[str]


class StringAnswerReq(CyanBaseModel):
    answer: str


class BoolAnswerReq(CyanBaseModel):
    answer: bool


AnswerReq = Union[StringArrayAnswerReq, StringAnswerReq, BoolAnswerReq]


def is_string_answer_req(a: AnswerReq) -> bool:
    return isinstance(a, StringAnswerReq)


def is_string_array_answer_req(a: AnswerReq) -> bool:
    return isinstance(a, StringArrayAnswerReq)


def is_bool_answer_req(a: AnswerReq) -> bool:
    return isinstance(a, BoolAnswerReq)




from dataclasses import dataclass
from typing import List, Dict, Any

from cyanprintsdk.domain.core.answer import Answer
from cyanprintsdk.domain.core.cyan import Cyan, CyanGlob


@dataclass
class CyanExtensionInput:
    prev_answers: List[Answer]
    prev_cyan: Cyan
    prev_extension_answers: Dict[str, List[Answer]]
    prev_extension_cyans: Dict[str, Cyan]


@dataclass
class CyanProcessorInput:
    read_dir: str
    write_dir: str
    globs: List[CyanGlob]
    config: Any


@dataclass
class CyanPluginInput:
    directory: str
    config: Any



import glob
import shutil
from pathlib import Path
from typing import List

from cyanprintsdk.domain.core.cyan import CyanGlob, GlobType
from cyanprintsdk.domain.core.fs.virtual_file import VirtualFile, VirtualFileReference


class CyanFileHelper:
    def __init__(self, read_dir: str, write_dir: str, globs: List[CyanGlob]):
        self._read_dir = read_dir
        self._write_dir = write_dir
        self.globs = globs

    @property
    def read_dir(self) -> str:
        return str(Path(self._read_dir).resolve())

    @property
    def write_dir(self) -> str:
        return str(Path(self._write_dir).resolve())

    def _glob_dir(self, g: CyanGlob) -> str:
        return str(Path(self.read_dir, g.root or ".").resolve())

    def _copy_file(self, from_path: str, to_path: str) -> None:
        parent = Path(to_path).parent
        parent.mkdir(parents=True, exist_ok=True)
        shutil.copy2(from_path, to_path)

    def resolve_all(self) -> List[VirtualFile]:
        copy = [x for x in self.globs if x.type.value == GlobType.Copy.value]
        template = [x for x in self.globs if x.type.value == GlobType.Template.value]

        for c in copy:
            self.copy(c)

        return [file for t in template for file in self.read(t)]

    def get(self, g: CyanGlob) -> List[VirtualFileReference]:
        gr: str = self._glob_dir(g)
        includes = set(glob.glob(g.glob, recursive=True, root_dir=gr))
        excludes = set.union(*[set(glob.glob(ex, recursive=True, root_dir=gr)) for ex in g.exclude])
        matched = [Path(gr, x) for x in list(includes - excludes)]

        return [VirtualFileReference(gr, self.write_dir, str(Path(x).relative_to(gr))) for x in matched]

    def read(self, g: CyanGlob) -> List[VirtualFile]:
        return [ref.read_file() for ref in self.get(g)]

    def copy(self, copy: CyanGlob) -> None:
        glob_root = self._glob_dir(copy)
        includes = set(glob.glob(copy.glob, recursive=True, root_dir=glob_root))
        excludes = set.union(*[set(glob.glob(ex, recursive=True, root_dir=glob_root)) for ex in copy.exclude])
        files = list(includes - excludes)

        for read in files:
            read_path = Path(glob_root, read)
            write_path = Path(self.write_dir, read_path.relative_to(glob_root))
            print(f"copy: {read_path} -> {write_path}")
            self._copy_file(str(read_path), str(write_path))
